from datetime import datetime, timedelta
import importlib.metadata
from pathlib import Path
from .grid import Grid
from .mpi import master_only, size

@master_only
def print_header():
    __version__ = importlib.metadata.version("elipy")
    # __version__ = "0.1.8"
    header_message = f"""
elipy v{__version__} -- post-processing tool for ABINIT EPH package
Started at: {datetime.now().strftime("%Y-%m-%d %H:%M:%S")}
    """
    print(header_message, flush=True)

@master_only
def print_mpi_info(num_kpoints: int) -> None:
    mpi_message = f"""
Number of cpu-s used is: {size}
K-point parallelization: avg. {num_kpoints} k-points per cpu
    """
    print(mpi_message, flush=True)

@master_only
def print_read_status(g_file: Path) -> None:
    read_status_message = f"""
Electron-phonon matrix elements, electron and phonon eigenvalues:\n{str(g_file)}
    """
    print(read_status_message, flush=True)
   
@master_only   
def print_variables(egrid: Grid, phgrid: Grid, e1grid: Grid = None) -> None:
    if e1grid:
        variables_message = f"""
    -------------------------------------------------------------------------------------
                        Variables that govern the present computation
    -------------------------------------------------------------------------------------

    all values in atomic units

    e_grid:
        e_window    {egrid.g_min:.3e}  {egrid.g_max:.3e}
        e_smearing  {egrid.smear:.3e} 
        e_npoints    {egrid.npoints}
    e1_grid:
        e1_window    {e1grid.g_min:.3e}  {e1grid.g_max:.3e}
        e1_smearing  {e1grid.smear:.3e} 
        e1_npoints    {e1grid.npoints}
    ph_grid:
        ph_window    {phgrid.g_min:.3e}  {phgrid.g_max:.3e}
        ph_smearing  {phgrid.smear:.3e} 
        ph_npoints    {phgrid.npoints}
    -------------------------------------------------------------------------------------

    -------------------------------------------------------------------------------------
        """
    else:
        variables_message = f"""
    -------------------------------------------------------------------------------------
                        Variables that govern the present computation
    -------------------------------------------------------------------------------------

    all values in atomic units

    e_grid:
        e_window    {egrid.g_min:.3e}  {egrid.g_max:.3e}
        e_smearing  {egrid.smear:.3e} 
        e_npoints    {egrid.npoints}
    ph_grid:
        ph_window    {phgrid.g_min:.3e}  {phgrid.g_max:.3e}
        ph_smearing  {phgrid.smear:.3e} 
        ph_npoints    {phgrid.npoints}
    -------------------------------------------------------------------------------------

    -------------------------------------------------------------------------------------
        """
        
    print(variables_message, flush=True)

@master_only
def print_computation() -> None:
    computation_message = """
-------------------------------------------------------------------------------------
                                Computation progress
-------------------------------------------------------------------------------------
    """
    print(computation_message, flush=True)

@master_only
def print_save_status(out_file: Path) -> None:
    save_status_message = f"""
Writing energy-resolved Eliashberg function values to netcdf file: {str(out_file)}
    """
    print(save_status_message, flush=True)
    
@master_only    
def print_complete(elapsed: float) -> None:
    complete_message = f"""
Calculation completed.
Calculation time is: {str(timedelta(seconds=elapsed))}
    """
    print(complete_message, flush=True)


import numpy as np
from numba import njit
from numba.typed import List

from .grid import Grid
from .kpt_utils import *
# from .mpi import master

@njit()
def gaussian(x: float, x0: float, sigma: float) -> np.float_:
    """gaussian Gaussian weight for given energy point x and eigenvalue x0

    Parameters
    ----------
    x : float
    x0 : float
    sigma : float

    Returns
    -------
    np.float_
    """    
    denom = 1/(2*sigma**2)
    pref = np.sqrt(denom/np.pi)
    return np.exp(-denom * (x-x0)**2) * pref

# @njit(parallel=True)
@njit()
def get_eew_term(eps_k: np.float_, eps_kq: np.float_, w_q: np.float_, 
                   sigma_eps: np.float_, sigma_eps_pr: np.float_, sigma_w: np.float_, 
                    eps_arr: np.ndarray, eps_pr_arr: np.ndarray, w_arr: np.ndarray,
                    Neps: int, Neps_pr: int, Nw: int, a2f_eew: np.ndarray) -> np.ndarray:
    """get_eew_term calculate gaussian weights for all eigenvalues

    Parameters
    ----------
    eps_k : np.float_
        electron eigenvalue at k-point
    eps_kq : np.float_
        electron eigenvalue at k+q-point
    w_q : np.float_
        phonon eigenvalue at q-point
    sigma_eps : np.float_
        electron smearing for e
    sigma_eps_pr : np.float_
        electron smearing for e'
    sigma_w : np.float_
        phonon smearing
    eps_arr : np.ndarray(Neps)
        energy grid e
    eps_pr_arr : np.ndarray(Neps_pr)
        energy grid e'
    w_arr : np.ndarray(Nw)
        frequency grid w
    Neps : int
        number of points in e grid
    Neps_pr : int
        number of points in e' grid
    Nw : int
        number of points in w grid
    a2f_eew : np.ndarray(Neps, Neps_pr, Nw)
        array for gaussian weights, being rewritten every time

    Returns
    -------
    np.ndarray
        gaussian weights
    """
    # we don't require Nw, may change later
    for i in range(Neps):
        for j in range(Neps_pr):
            a2f_eew[i,j,:] = ( gaussian(w_arr[:], w_q, sigma_w) 
                            * gaussian(eps_pr_arr[j], eps_kq, sigma_eps_pr)
                            * gaussian(eps_arr[i], eps_k, sigma_eps) )
    return a2f_eew

@njit()
def get_ew_term(eps_k: np.float_, eps_kq: np.float_, w_q: np.float_, 
                   sigma_eps: np.float_, sigma_w: np.float_, 
                    eps_arr: np.ndarray, w_arr: np.ndarray,
                    Neps: int, Nw: int, a2f_ew: np.ndarray) -> np.ndarray:
    """get_ew_term calculate gaussian weights for all eigenvalues

    Parameters
    ----------
    eps_k : np.float_
        electron eigenvalue at k-point
    eps_kq : np.float_
        electron eigenvalue at k+q-point
    w_q : np.float_
        phonon eigenvalue at q-point
    sigma_eps : np.float_
        electron smearing for e
    sigma_w : np.float_
        phonon smearing
    eps_arr : np.ndarray(Neps)
        energy grid e
    w_arr : np.ndarray(Nw)
        frequency grid w
    Neps : int
        number of points in e grid
    Nw : int
        number of points in w grid
    a2f_ew : np.ndarray(Neps, Nw)
        array for gaussian weights, being rewritten every time

    Returns
    -------
    np.ndarray
        gaussian weights
    """
    # we don't require Nw, may change later
    for i in range(Neps):
        a2f_ew[i,:] = ( gaussian(w_arr[:], w_q, sigma_w) 
                        * gaussian(eps_arr[i], eps_kq, sigma_eps)
                        * gaussian(eps_arr[i], eps_k, sigma_eps) )
    return a2f_ew

@njit()
def calculate_chunk(gkq_chunk: np.ndarray, eps_eig: np.ndarray, ph_eig: np.ndarray, 
                    nonzero_dims: tuple, ikqpts: list,
                    egrid: np.ndarray, e1grid: np.ndarray, phgrid: np.ndarray,
                    esmear: np.float_, e1smear: np.float_, phsmear: np.float_,
                    epoints: int, e1points: int, phpoints: int) -> np.ndarray:
    """calculate_chunk calculate a2F values for given |g|^2 chunk

    Parameters
    ----------
    gkq_chunk : np.ndarray
        chunk of |g|^2 values
    eps_eig : np.ndarray
        electron eigenvalues
    ph_eig : np.ndarray
        phonon eigenvalues
    nonzero_dims : tuple
        indicies where |g|^2 != 0
    ikqpts : list
        maps of k->k+q points for every q
    egrid : np.ndarray
        energy grid e
    e1grid : np.ndarray
        energy grid e'
    phgrid : np.ndarray
        frequency grid w
    esmear : np.float_
        electron smearing for e
    e1smear : np.float_
        electron smearing for e'
    phsmear : np.float_
        phonon smearing
    epoints : int
        number of points in e grid
    e1points : int
        number of points in e' grid
    phpoints : int
        number of points in w grid
    Returns
    -------
    np.ndarray
        a2F values for |g|^2 chunk
    """
    # countainers for a2f values
    a2f_vals = np.zeros((epoints, e1points, phpoints))
    a2f_temp = np.empty((epoints, e1points, phpoints))
    ewindow_bounds = [np.amin(egrid), np.amax(egrid)]
    e1window_bounds = [np.amin(e1grid), np.amax(e1grid)]    
    nk, nq, nnu, nb, nbp = nonzero_dims
    for i in range(len(nk)):
        ik, iq, inu, ib, ibp = nk[i], nq[i], nnu[i], nb[i], nbp[i]
        ikq = ikqpts[iq][ik][0]
        w_q = ph_eig[iq, inu]
        eps_k = eps_eig[ik,ib]
        if (eps_k < ewindow_bounds[0])|(eps_k > ewindow_bounds[1]):
            continue # no need to account these values 
        eps_kq = eps_eig[ikq,ibp]
        if (eps_kq < e1window_bounds[0])|(eps_kq > e1window_bounds[1]):
            continue # no need to account these values 
        g_kq = gkq_chunk[ik,iq,inu,ib,ibp]
        get_eew_term(eps_k, eps_kq, w_q,
                esmear, e1smear, phsmear,
                egrid, e1grid, phgrid,
                epoints, e1points, phpoints,
                a2f_temp)
        a2f_vals += a2f_temp * g_kq
    return a2f_vals

@njit()
def calculate_reduced_chunk(
    gkq_chunk: np.ndarray, eps_eig: np.ndarray, ph_eig: np.ndarray, 
    nonzero_dims: tuple, ikqpts: list,
    egrid: np.ndarray, phgrid: np.ndarray,
    esmear: np.float_, phsmear: np.float_,
    epoints: int, phpoints: int
) -> np.ndarray:
    """calculate_chunk calculate a2F values for given |g|^2 chunk
    for the special case e' == e

    Parameters
    ----------
    gkq_chunk : np.ndarray
        chunk of |g|^2 values
    eps_eig : np.ndarray
        electron eigenvalues
    ph_eig : np.ndarray
        phonon eigenvalues
    nonzero_dims : tuple
        indicies where |g|^2 != 0
    ikqpts : list
        maps of k->k+q points for every q
    egrid : np.ndarray
        energy grid e
    phgrid : np.ndarray
        frequency grid w
    esmear : np.float_
        electron smearing for e
    phsmear : np.float_
        phonon smearing
    epoints : int
        number of points in e grid
    phpoints : int
        number of points in w grid
    Returns
    -------
    np.ndarray
        a2F values for |g|^2 chunk
    """
    # countainers for a2f values
    a2f_vals = np.zeros((epoints, phpoints))
    a2f_temp = np.empty((epoints, phpoints))
    ewindow_bounds = [np.amin(egrid), np.amax(egrid)]
    nk, nq, nnu, nb, nbp = nonzero_dims
    for i in range(len(nk)):
        ik, iq, inu, ib, ibp = nk[i], nq[i], nnu[i], nb[i], nbp[i]
        ikq = ikqpts[iq][ik][0]
        w_q = ph_eig[iq, inu]
        eps_k = eps_eig[ik,ib]
        eps_kq = eps_eig[ikq,ibp]
        if (eps_k < ewindow_bounds[0])|(eps_k > ewindow_bounds[1]):
            continue # no need to account these values 
        if (eps_kq < ewindow_bounds[0])|(eps_kq > ewindow_bounds[1]):
            continue # no need to account these values 
        g_kq = gkq_chunk[ik,iq,inu,ib,ibp]
        get_ew_term(eps_k, eps_kq, w_q,
                esmear, phsmear,
                egrid, phgrid,
                epoints, phpoints,
                a2f_temp)
        a2f_vals += a2f_temp * g_kq
    return a2f_vals
    
# def get_a2f_chunk(gkq_chunk: np.ndarray, kpts: np.ndarray, kpts_chunk: np.ndarray, qpts: np.ndarray,
#                   eps_eig: np.ndarray, ph_eig: np.ndarray, 
#                   egrid: Grid, e1grid: Grid, phgrid: Grid) -> np.ndarray:
#     """get_a2f_chunk wrapper for calculate_chunk function

#     Parameters
#     ----------
#     gkq_chunk : np.ndarray
#         chunk of |g|^2 values
#     kpts : np.ndarray
#         k-points in full BZ
#     kpts_chunk : np.ndarray
#         chunk of k-points
#     qpts : np.ndarray
#         q-points in full BZ
#     eps_eig : np.ndarray
#         electron eigenvalues
#     ph_eig : np.ndarray
#         phonon eigenvalues
#     egrid : Grid
#         electron grid e
#     e1grid : Grid
#         electron grid e'
#     phgrid : Grid
#         phonon grid w

#     Returns
#     -------
#     np.ndarray
#         a2F values for |g|^2 chunk
#     """
#     # mapping of BZ: for every q get k+q -> k 
#     ikqpts = List()
#     [ikqpts.append(get_kq2k(kpts, kpts_chunk, qpt)) for qpt in qpts]
#     # take only nonzero values of gkq_chunk
#     where_nonzero = np.nonzero(gkq_chunk)
#     a2f_vals = calculate_chunk(gkq_chunk, eps_eig, ph_eig,
#                                where_nonzero, ikqpts, 
#                                egrid.grid, e1grid.grid, phgrid.grid,
#                                egrid.smear, e1grid.smear, phgrid.smear,
#                                egrid.npoints, e1grid.npoints, phgrid.npoints)
#     return a2f_vals



# some useful constants
from scipy.constants import physical_constants as pc

Ha_eV = pc['hartree-electron volt relationship'][0]
eV_Ha = pc['electron volt-hartree relationship'][0]
THz_Ha = pc['hertz-hartree relationship'][0]*1e9
inv_cm_Ha = pc['inverse meter-hartree relationship'][0]*1e2

unit_conversion = {
    'Ha': 1e0,
    'eV': eV_Ha,
    'Thz': THz_Ha,
    'cm-1': inv_cm_Ha,
    'meV': eV_Ha*1e-3,
}

# egird defaults in Ha units
default_emin = -0.4
default_emax = 0.4
default_esmear = 0.01 # Ha
default_epoints = 50

# phgrid defaults in meV
# default min and max are taken from eigenfrequencies file
default_phsmear = 2e-5 # ~0.544 meV
default_phpoints = 100 # higher resolution over phonon frequencies

# small addition to highest phonon eigenvalue for phonon grid
ph_delta = eV_Ha * 1e-2 # 10 meV



"""This file is taken from https://github.com/GkAntonius/ElectronPhononCoupling/blob/master/ElectronPhononCoupling"""

"""MPI distribution of work load"""
import sys, traceback
from contextlib import contextmanager
from mpi4py import MPI

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

master = bool(rank == 0)

@contextmanager
def mpi_abort_if_exception():
    """mpi_abort_if_exception terminates all mpi process if an exception is raised.
    """
    try:
        yield
    except:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        traceback.print_exception(exc_type, exc_value, exc_traceback)
        MPI.COMM_WORLD.Abort(1)

def mpi_watch(f):
    """mpi_watch is decorator. Terminates all mpi process if an exception is raised.

    Parameters
    ----------
    f : function

    Returns
    -------
    function
    """
    def g(*args, **kwargs):
        with mpi_abort_if_exception():
            return f(*args, **kwargs)
    return g

def master_only(f):
    """master_only is ecorator. Lets a function be executed only by master.

    Parameters
    ----------
    f : function

    Returns
    -------
    function
    """
    def g(*args, **kwargs):
        if master:
            with mpi_abort_if_exception():
                return f(*args, **kwargs)
        return
    return g


from pathlib import Path
from time import time

import numpy as np
from numba.typed import List

from .files_handling import *
from .constants import *
from .messages import *
from .grid import Grid
from .a2f_calculator import calculate_chunk, calculate_reduced_chunk #, get_a2f_chunk 
from .kpt_utils import get_kq2k

from .mpi import MPI, comm, size, rank, master_only, mpi_watch, master

class Eliashberg:
    """ contains methods for Eliashberg function a2F(e,e',w) calculation
    """
    def __init__(self, gkq_file: str|Path,
                 ee1grid: bool = False,
                 ewindow: list = [default_emin, default_emax], esmear: float = default_esmear, 
                 epoints: int = default_epoints, eunits: str = 'Ha',
                 e1window: list = None, e1smear: float = None,
                 e1points: int = None, e1units: str = None,
                 phwindow: list = None, phsmear: float = default_phsmear,
                 phpoints: int = default_phpoints, phunits: str = 'Ha',
                 out_file: str|Path = Path.cwd()/'eliashberg_eew.nc') -> None:
        """__init__ parameters for Eliashberg class

        Parameters
        ----------
        gkq_file : str | Path
            Path to netCDF4 file with |g|^2 eph matrix elements
        ee1grid: bool, optional
            If true, code calculates a2F(e,e',w); otherwise, a2F(e,w) is calculated
        ewindow : list, optional
            window for electron energies,
            by default [default_emin, default_emax]
        esmear : float, optional
            smearing of electron delta function,
            by default default_esmear
        epoints : int, optional
            number of points in electron energy grid,
            by default default_epoints
        eunits : str, optional
            electron energy unit, by default 'Ha'
        e1window : list, optional
            same as `ewindow` but for e',
            by default None (taken from ewindow)
        e1smear : float, optional
            same as `esmear` but for e',
            by default None (taken from esmear)
        e1points : int, optional
            same as `epoints` but for e',
            by default None (taken from epoints)
        e1units : str, optional
            same as `eunits` but for e',
            by default None (taken from eunits)
        phwindow : list, optional
            window for phonon frequencies,
            by default None (taken from phonon eigenvalues)
        phsmear : float, optional
            smearing of phonon delta function,
            by default default_phsmear
        phpoints : int, optional
            number of points in phonon frequency grid,
            by default default_phpoints
        phunits : str, optional
            phonon frequency unit, by default 'Ha'
        out_file : str | Path, optional
            path to netCDF4 file for output storage,
            by default Path.cwd()/'eliashberg_eew.nc'
        """
        self.gkq_file = gkq_file
        self.ee1grid = ee1grid
        self.egrid = Grid(*ewindow, esmear, epoints, eunits)
        # use the same parameters for e and e1 by default
        if self.ee1grid:
            e1window = ewindow if not e1window else e1window
            e1smear = esmear if not e1smear else e1smear
            e1points = epoints if not e1points else e1points
            e1units = eunits if not e1units else e1units
            self.e1grid = Grid(*e1window, e1smear, e1points, e1units)
        # phonon grid is set according to phonon eigenvalues by default. 
        # If phonon window is chosen, use these values
        if phwindow:
            self.phgrid = Grid(*phwindow, phsmear, phpoints, phunits)
        else:
            self.phgrid = None # will be initialized later
        self.__phsmear = phsmear
        self.__phpoints = phpoints
        self.__phunits = phunits
        self.out_file = out_file
    
    @master_only        
    def read_data(self):
        """read_data reads input files and check their consistency
        """
        # read gkq file
        gkq_data = get_gkq_values(self.gkq_file)
        gkq_vals = gkq_data['gkq']
        self.g_kpts, self.g_qpts, self.efermi = gkq_data['kpts'], gkq_data['qpts'], gkq_data['efermi']
        self.e_eigvals = gkq_data['e_eigs']
        self.ph_eigvals = gkq_data['ph_eigs']
        # if everything is OK, make electron eigenvalues Fermi energy centered
        self.e_eigvals = self.e_eigvals - self.efermi
        # g_kq has shape `[nq,nk,...]`, we swap 0th and 1st dimensions for convenience
        self.gkq_vals = np.swapaxes(gkq_vals, 0, 1)
        # ensure gkq array to be C-contiguous in memory
        self.gkq_vals = np.ascontiguousarray(self.gkq_vals)
        # and set up variables for all dimensions
       
    @mpi_watch
    def broadcast_dims(self):
        """broadcast_dims broadcasts dimensions of all arrays to cpus
        """               
        if master:
            dim_arr = np.array(self.gkq_vals.shape[:-1], dtype=int)
        else:
            dim_arr = np.empty(4, dtype=int)
        comm.Bcast([dim_arr, MPI.INT])
        self.nkpt, self.nqpt, self.nbranch, self.nband = dim_arr
    
    @mpi_watch
    def broadcast_kqpoints(self):
        """broadcast_kqpoints broadcasts k- and q-point arrays to cpus
        """       
        if not master:
            self.g_kpts = np.empty((self.nkpt, 3), dtype=np.float64)
            self.g_qpts = np.empty((self.nqpt, 3), dtype=np.float64)
            # self.ewin_ikpts = None
            # self.e1win_ikpts = None
        comm.Bcast([self.g_kpts, MPI.DOUBLE])
        comm.Bcast([self.g_qpts, MPI.DOUBLE])
        
    @mpi_watch
    def broadcast_eigvals(self):
        """broadcast_eigvals broadcasts electron and phonon eigenvalues to cpus
        """
        if not master:
            self.e_eigvals = np.empty((self.nkpt, self.nband), dtype=np.float64)
            self.ph_eigvals = np.empty((self.nqpt, self.nbranch), dtype=np.float64)
        comm.Bcast([self.e_eigvals, MPI.DOUBLE])
        comm.Bcast([self.ph_eigvals, MPI.DOUBLE])
        # if phonon grid is not set, do it now    
        phmin = np.amin(self.ph_eigvals)
        phmin = 0 if phmin >= 0 else phmin - ph_delta
        if not self.phgrid:           
            self.phgrid = Grid(phmin, np.amax(self.ph_eigvals) + ph_delta,
                               self.__phsmear*unit_conversion[self.__phunits],
                               self.__phpoints, 'Ha')


    # here array is flatten to 1d and back to nd, but it may brake everything at some point
    # @mpi_watch
    # def scatter_gkq_vals(self):
    #     """scatter_gkq_vals distributes |g|^2 values over cpus
    #     """
    #     # TODO: think about scattering over nqpt as well
    #     if master: 
    #         gkq_flatten = self.gkq_vals.flatten()
    #     else:
    #         gkq_flatten = None
    #     # find dimension of chunks per cpu and number dimension displacemen
    #     ave, res = divmod(self.nkpt, size)
    #     counts = np.array([ave + 1 if p < res else ave for p in range(size)], dtype=int)
    #     # self because will call it later
    #     self.__displ = np.array([sum(counts[:p]) for p in range(size)], dtype=int)
    #     rest_dims = self.nqpt*self.nbranch*self.nband*self.nband
    #     # allocate space for chunk
    #     self.gkq_chunk = np.empty(counts[rank]*rest_dims)
    #     comm.Scatterv([gkq_flatten, counts*rest_dims, self.__displ*rest_dims, MPI.DOUBLE],
    #                   self.gkq_chunk, root=0)    
    #     self.gkq_chunk = self.gkq_chunk.reshape((counts[rank], self.nqpt, self.nbranch,
    #                                              self.nband, self.nband))

    # TODO: think how to send LARGE amounts of data
    @mpi_watch
    def scatter_gkq_vals(self):
        """scatter_gkq_vals distributes |g|^2 values over cpus
        """
        # find dimension of chunks per cpu and number dimension displacemen
        ave, res = divmod(self.nkpt, size)
        counts = np.array([ave + 1 if p < res else ave for p in range(size)], dtype=int)
        self.__displ = np.array([sum(counts[:p]) for p in range(size)], dtype=int)
        if master:
            for i in range(1, size):
                if i != size-1:
                    comm.Send([self.gkq_vals[self.__displ[i]:self.__displ[i+1]], MPI.REAL8],
                            dest=i, tag=10+i)
                else:
                    comm.Send([self.gkq_vals[self.__displ[i]:], MPI.REAL8],
                            dest=i, tag=10+i)
            self.gkq_chunk = self.gkq_vals[:self.__displ[1]]
        else:
            chunk_shape = (counts[rank],self.nqpt,self.nbranch,self.nband,self.nband)
            self.gkq_chunk = np.empty(chunk_shape, dtype=np.float64)
            comm.Recv([self.gkq_chunk, MPI.REAL8], source=0, tag=10+rank)
        
    @mpi_watch
    def sum_chunks(self):
        """sum_chunks calculates a2f values for distributed |g|^2 chunk and sums them
        """
        # allocate array summing chunks from all cpu-s
        if rank == 0:
            if self.ee1grid:
                self.a2f_vals = np.empty((self.egrid.npoints,
                                    self.e1grid.npoints,
                                    self.phgrid.npoints)) 
            else:
                self.a2f_vals = np.empty((self.egrid.npoints,
                                        self.phgrid.npoints)) 
        else:
            self.a2f_vals = None
        
        # get the proper subset of kpoints
        if rank != size-1:
            kpts_chunk = self.g_kpts[self.__displ[rank]:self.__displ[rank+1],:]
        else:
            kpts_chunk = self.g_kpts[self.__displ[rank]:,:]

        # mapping of BZ: for every q get k+q -> k 
        ikqpts = List()
        [ikqpts.append(get_kq2k(self.g_kpts, kpts_chunk, qpt)) for qpt in self.g_qpts]
        # take only nonzero values of gkq_chunk
        where_nonzero = np.nonzero(self.gkq_chunk)
        if self.ee1grid:
            # calculate one chunk
            a2f_chunk = calculate_chunk(
                self.gkq_chunk, self.e_eigvals, self.ph_eigvals,
                where_nonzero, ikqpts,
                self.egrid.grid, self.e1grid.grid, self.phgrid.grid,
                self.egrid.smear, self.e1grid.smear, self.phgrid.smear,
                self.egrid.npoints, self.e1grid.npoints, self.phgrid.npoints
                                        )
        else:
            a2f_chunk = calculate_reduced_chunk(
                self.gkq_chunk, self.e_eigvals, self.ph_eigvals,
                where_nonzero, ikqpts,
                self.egrid.grid, self.phgrid.grid,
                self.egrid.smear, self.phgrid.smear,
                self.egrid.npoints, self.phgrid.npoints
            )

        # calculate a2f values for given chunck        
        # a2f_chunk = get_a2f_chunk(self.gkq_chunk, self.g_kpts, kpts_chunk, self.g_qpts, 
        #                           self.e_eigvals, self.ph_eigvals,
        #                           self.egrid, self.e1grid, self.phgrid)
        
        comm.Reduce([a2f_chunk, MPI.DOUBLE], [self.a2f_vals, MPI.DOUBLE],
                    op=MPI.SUM, root=0)       
            
    @master_only
    def write_netcdf(self):
        """write_netcdf writes a2f to netCDF4 file
        """
        # multiply Eliashberg function to (almost) correct prefactor
        self.a2f_vals = self.a2f_vals * (2/self.nkpt/self.nqpt) # division to g(Ef) is up to user!
        # save computed Eliashberg function to netCDF file
        if self.ee1grid:
            store_a2f_values(self.out_file, self.efermi, 
                            self.egrid, self.e1grid, self.phgrid, self.a2f_vals)
        else:
            store_a2f_reduced_values(self.out_file, self.efermi, 
                            self.egrid, self.phgrid, self.a2f_vals)
                    
    def compute_a2f(self):
        """compute_a2f gathers all previous methods together into united workflow
        """
        # and add status messages to log file
        if master:
        # we don't need to log everything 
            start = time()           
        print_header()
        self.read_data()
        print_read_status(self.gkq_file)
        self.broadcast_dims()
        print_mpi_info(self.nkpt//size)
        self.broadcast_kqpoints()
        self.broadcast_eigvals()
        self.scatter_gkq_vals()
        if self.ee1grid:
            print_variables(self.egrid, self.phgrid, self.e1grid)
        else:
            print_variables(self.egrid, self.phgrid)
        
        # print_computation()
        self.sum_chunks()
        
        self.write_netcdf()
        print_save_status(self.out_file)
        
        if master:
            elapsed = time() - start
            print_complete(elapsed)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['elipy', 'elipy.core']

package_data = \
{'': ['*']}

install_requires = \
['mpi4py>=3.1.4,<4.0.0',
 'netCDF4>=1.5.7,<2.0.0',
 'numba>=0.56.4,<0.57.0',
 'numpy>=1.23.5,<2.0.0',
 'scipy>=1.9.3,<2.0.0']

setup_kwargs = {
    'name': 'elipy',
    'version': '0.1.8',
    'description': 'Post-processing tool for Abinit >= 9.8.1 aiming to calculate energy-resolved Eliashberg function',
    'long_description': "# Elipy\n\nPost-processing tool for ABINIT >= 9.8.1 aiming for calculation of energy-resolved Eliashberg function\n\n## Description\n\nSince version 9.8.1 ABINIT provides easy-to-use methods within EPH package for the calculation of electron-phonon matrix elements on rectangular k- and q-point grids (optdriver 7, eph_task 11). The present package is a tool for calculation of electron-energy-resolved Eliashberg function on arbitrary electron and phonon grids. \n\n$$\n\\alpha^2 F(\\epsilon,\\epsilon',\\omega)=\\frac{1}{N_\\mathbf{k}N_\\mathbf{q}}\\sum_{\\mathbf{k}\\mathbf{q}mn\\nu}\\left|g_{mn\\nu}(\\mathbf{k},\\mathbf{q})\\right|^2\\delta(\\epsilon_{\\mathbf{k}n}-\\epsilon)\\delta(\\epsilon_{\\mathbf{k+q}m}-\\epsilon')\\delta(\\omega_{\\mathbf{q}\\nu}-\\omega)\n$$\n\nThe above-written definition does not include electron density of states at Fermi level. One needs to divide $\\alpha^2 F$ over $N_F$ to get physically meaningful results. \n\nProject uses mpi4py for many-core parallelization and Numba for acceleration of procedures dealing with iteration over numpy arrays.\n\nElipy is insipired by [ElectronPhononCoupling](https://github.com/GkAntonius/ElectronPhononCoupling) and [Abipy](https://github.com/abinit/abipy) projects, and uses some of their utility functions. Apologies for code copy-pasting, it allows to keep dependency list as short as possible.  \n\n### Limitations\n\n* only Gaussian smearing summation for delta functions\n* no account for spin of electron states\n\n## Getting Started\n\n### Dependencies\n\n* Numpy\n* Scipy\n* netCDF4\n* Numba\n* mpi4py\n\nThe actual versions of required packages are stored in pyproject.toml file.\n\n### Installation\n\n```\npip install elipy\n```\nALternatively, one can build code with [Poetry](https://python-poetry.org) package manager:\n\n```\ngit clone https://github.com/Radioteddy/elipy.git\ncd elipy\npoetry install\n```\n### Executing program\n\n* Use mpirun, mpiexec, srun,... for program execution\n```\nmpiexec -n X python filename.py > out 2> err\n```\n\n## Authors\n\n[Fedor Akhmetov](https://github.com/Radioteddy)\n\n## Version History\n\n* 0.1.8\n    * collective communication for matrix elements replaced by p2p one\n* 0.1.7\n    * electron and phonon eigenvalues are taken from GSTORE\n* 0.1.6\n    * minor bugfixes and improvements\n* 0.1.5\n    * Working Release\n* 0.1.0\n    * Initial Release\n\n## Acknowledgments\n* [ElectronPhononCoupling](https://github.com/GkAntonius/ElectronPhononCoupling)\n* [Abipy](https://github.com/abinit/abipy)\n",
    'author': 'Fedor Akhmetov',
    'author_email': 'f.akhmetov@utwente.nl',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path

import numpy as np
import netCDF4 as nc

from .grid import Grid
from .kpt_utils import get_eigvals_bz

def check_in_path(path: str|Path, file_type: str = None) -> Path:
    """check_in_path converts input file path to pathlib.Path if necessary and makes it absolute

    Parameters
    ----------
    path : str | Path
        file path
    file_type : str
        which file

    Returns
    -------
    Path
        absolute path to file
    """    
    # pathlib.Path representation of input file
    if isinstance(path, Path):
        pass
    elif isinstance(path, str):
        path = Path(path)
    else:
        raise TypeError('file_path should be str or pathlib.Path object')    
    # check if file really exists and make it absolute
    if not path.is_absolute():
        path = Path.cwd() / path
    if not path.exists():
        raise FileNotFoundError(f"Path to {file_type} file '{str(path)}' does not exist")
    return path
    
def check_out_path(path: str|Path) -> Path:
    """check_out_path converts output file path to pathlib.Path if necessary and makes it absolute

    Parameters
    ----------
    path : str | Path
        file path

    Returns
    -------
    Path
        absolute path to file
    """
    if isinstance(path, Path):
        pass
    elif isinstance(path, str):
        path = Path(path)
    else:
        raise TypeError('file_path should be str or pathlib.Path object')
    # check if directory to output file exists, and create it if necessary    
    if not path.parents[0].exists():
        path.parents[0].mkdir(parents=True, exist_ok=True)
    return path
                
def get_electron_eigenvalues_kpoints(eigenval_path: str|Path) -> tuple([np.ndarray, np.ndarray]):
    """get_electron_eigenvalues_kpoints reads netCDF4 file with electron eigenvalues

    Parameters
    ----------
    eigenval_path : str | Path
        path to eigenvalues file

    Returns
    -------
    (numpy.ndarray, numpy.ndarray)
        arrays of electron eigenvalues and k-points
    """
    eigenval_path = check_in_path(eigenval_path, 'Electron eigenvalues')
    if 'EIG' in eigenval_path.name:
        file=nc.Dataset(eigenval_path, 'r')
        eigenvalues = np.ma.getdata(file.variables['Eigenvalues'][:])[0,:,:]
        kpoints = np.ma.getdata(file.variables['Kptns'][:])
        file.close()
    elif ('GSTORE' in eigenval_path.name) or ('A2F' in eigenval_path.name):
        file=nc.Dataset(eigenval_path, 'r')
        # currently for spin-independent eigenvalues only
        eigenvalues = np.ma.getdata(file.variables['eigenvalues'][:])[0,:,:]
        kpoints = np.ma.getdata(file.variables['reduced_coordinates_of_kpoints'][:])
        file.close()
    else:
        raise ValueError('incorrect netCDF file!')
    return eigenvalues, kpoints

def get_phonon_eigenvalues(eigenval_path: str|Path) -> np.ndarray:
    """get_phonon_eigenvalues reads acsii file with phonon eigenvalues

    Parameters
    ----------
    eigenval_path : str | Path
        path to eigenvalues file

    Returns
    -------
    numpy.ndarray
        array of phonon eigenvalues
    """    
    eigenval_path = check_in_path(eigenval_path, 'Phonon eigenvalues')
    if 'PHFRQ' in eigenval_path.name:
        eigenvalues = np.loadtxt(eigenval_path, usecols=(1,2,3))
    else:
        raise ValueError('incorrect phonon modes file!')
    return eigenvalues

def get_gkq_values(gkq_path: str) -> dict:
    """get_gkq_values reads netCDF4 file with eph matrix elements

    Parameters
    ----------
    gkq_path : str
        path to |g|^2 file 

    Returns
    -------
    (numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.float_)
        arrays with |g|^2 values, k- and q-points, Fermi energy
    """    
    gkq_path = check_in_path(gkq_path, 'EPH matrix elements')
    read_data = {}
    if 'GSTORE' in gkq_path.name:
        file = nc.Dataset(gkq_path, 'r')
        # read electron eigenenergies in IBZ and restore values for full BZ
        k_ibz2bz = np.ma.getdata(file.variables['gstore_kbz2ibz'][:])
        e_eigs_ibz = np.ma.getdata(file.variables['eigenvalues'][:])[0,:,:]
        e_eigs_bz = get_eigvals_bz(e_eigs_ibz, k_ibz2bz)
        read_data['e_eigs'] = e_eigs_bz
        # read phonon frequencies in IBZ and restore values for full BZ
        q_ibz2bz = np.ma.getdata(file.variables['gstore_qbz2ibz'][:])
        ph_eigs_ibz = np.ma.getdata(file.variables['phfreqs_ibz'][:])
        ph_eigs_bz = get_eigvals_bz(ph_eigs_ibz, q_ibz2bz)
        read_data['ph_eigs'] = ph_eigs_bz
        if len(file.groups) == 1:
            gkq = file.groups['gqk_spin1'].variables['gvals']
            gkq_vals = np.ma.getdata(gkq[:])
            # check if we got g instead of |g|^2 and calculate |g|^2
            if gkq_vals.shape[-1] == 2:
                gkq_vals_temp = np.empty(gkq_vals.shape[:-1])
                gkq_vals_temp = gkq_vals[:,:,:,:,:,0]**2 + gkq_vals[:,:,:,:,:,1]**2
                gkq_vals = gkq_vals_temp
            else:
                # since we deal with |g|^2 we don't need last dimension
                gkq_vals = gkq_vals.reshape(gkq_vals.shape[:-1]) 
            kpoints = np.ma.getdata(file.variables['gstore_kbz'][:])
            qpoints = np.ma.getdata(file.variables['gstore_qbz'][:])
            fermi_energy = np.float64(np.ma.getdata(file.variables['fermi_energy'][:]))
            assert (gkq_vals.shape[0], gkq_vals.shape[1]) == (qpoints.shape[0], kpoints.shape[0]), "only k(bz) q(bz) gkq file is currently supported"
            read_data['kpts'] = kpoints
            read_data['qpts'] = qpoints
            read_data['gkq'] = gkq_vals
            read_data['efermi'] = fermi_energy
        else:
            raise ValueError('Spin-dependent gkq is not supported yet')
        file.close()
    else:
        raise ValueError('incorrect netCDF file!')
    return read_data

def store_a2f_values(out_path: str|Path, efermi: np.float_,
                     egrid: Grid, e1grid: Grid, phgrid: Grid, a2f_vals: np.ndarray) -> None:
    """store_a2f_values saves output data no netCDF4 file

    Parameters
    ----------
    out_path : str | Path
        path to output file
    efermi : np.float_
        Fermi energy of a system
    egrid : Grid
        electron energy grid e
    e1grid : Grid
        electron energy grid e'
    phgrid : Grid
        phonon frequency grid w
    a2f_vals : np.ndarray
        a2F(e,e',w) values
    """
    out_path = check_out_path(out_path)
    ncout = nc.Dataset(out_path,'w')
    efvar = ncout.createVariable('Fermi_energy', 'float64') 
    efvar[0] = efermi
    ncout.createDimension('number_of_epoints', egrid.npoints)
    ncout.createDimension('number_of_e1points',e1grid.npoints)
    ncout.createDimension('number_of_frequencies', phgrid.npoints)
    evar = ncout.createVariable('energy','float64',('number_of_epoints'))
    evar[:] = egrid.grid
    e1var = ncout.createVariable('energy_pr','float64',('number_of_e1points'))
    e1var[:] = e1grid.grid
    phvar = ncout.createVariable('frequency','float64',('number_of_frequencies'))
    phvar[:] = phgrid.grid
    a2fvar = ncout.createVariable('a2f','float64',('number_of_epoints',
                                                   'number_of_e1points','number_of_frequencies'))
    a2fvar[:] = a2f_vals
    ncout.close()      

def store_a2f_reduced_values(out_path: str|Path, efermi: np.float_,
                     egrid: Grid, phgrid: Grid, a2f_vals: np.ndarray) -> None:
    """store_a2f_values saves output data no netCDF4 file
    for the special case e' == e

    Parameters
    ----------
    out_path : str | Path
        path to output file
    efermi : np.float_
        Fermi energy of a system
    egrid : Grid
        electron energy grid e
    phgrid : Grid
        phonon frequency grid w
    a2f_vals : np.ndarray
        a2F(e,w) values
    """
    out_path = check_out_path(out_path)
    ncout = nc.Dataset(out_path,'w')
    efvar = ncout.createVariable('Fermi_energy', 'float64') 
    efvar[0] = efermi
    ncout.createDimension('number_of_epoints', egrid.npoints)
    ncout.createDimension('number_of_frequencies', phgrid.npoints)
    evar = ncout.createVariable('energy','float64',('number_of_epoints'))
    evar[:] = egrid.grid
    phvar = ncout.createVariable('frequency','float64',('number_of_frequencies'))
    phvar[:] = phgrid.grid
    a2fvar = ncout.createVariable('a2f','float64',
                                  ('number_of_epoints','number_of_frequencies'))
    a2fvar[:] = a2f_vals
    ncout.close()      


from numpy import linspace
from .constants import *

class Grid:
    """ provides linear grid for electron and phonon energies
    """    
    def __init__(self, g_min: float, g_max: float, smear: float,
                 npoints: int, unit: str = 'Ha'):
        """__init__ parameters of grid

        Parameters
        ----------
        g_min : float
            lowest grid point
        g_max : float
            largest grid point
        smear : float
            parameter for gaussian smearing
        npoints : int
            number of grid points
        unit : str, optional
           energy units, by default 'Ha'
        """
        self.unit = unit
        self.npoints = npoints
        if self.unit == 'Ha':
            self.g_min = g_min
            self.g_max = g_max
            self.smear = smear
        elif self.unit in unit_conversion.keys():       
            self.g_min = g_min * unit_conversion[self.unit]
            self.g_max = g_max * unit_conversion[self.unit]
            self.smear = smear * unit_conversion[self.unit]
        else:
            raise ValueError('unit is not allowed')
        self.grid = self.make_grid()
                
    def make_grid(self):
        return linspace(self.g_min, self.g_max, self.npoints)


"""
elipy: a post-processing tool for ABINIT EPH module

Provides `Eliashberg` class object contaning methods for calculation of energy-dependent Eliashberg function a2F from e-ph matrix elements |g_kq|^2

Usage: 
1. Create file with input parameters `yourfile.py`, e.g.
-----------------------------------------------------------
from elipy import Eliashberg

gkq_input = 'indata\in_GSTORE.nc'
eeig_input = 'indata\in_EIG.nc'
pheig_input = 'indata\in_PHFRQ'

a2f_class = Eliashberg(gkq_input, eeig_input, pheig_input)
a2f_class.compute_a2f()
-----------------------------------------------------------
2. Run the caculation in the following way: `mpirun -n 4 python yourfile.py > log`
3. At the end of calculation you will get netCDF4 file `eliashberg_eew.nc' with a2F(e,e',w)
"""

from .core.eliashberg import *
from .core import kpt_utils as kpt_utils
from .core import grid as grid
from .core import constants as constants
from .core import a2f_calculator as a2f_calculator


# These functions are taken from abipy package (almost) without any changes
# https://abinit.github.io/abipy/_modules/abipy/core/kpoints.html
# the reason to take them instead of import is keeping minimalistic requirements
# and accelerate up to 2-3 orders of magnitude thanks to Numba jit-compilation
import numpy as np
from numba import njit

@njit()
def allclose(x: np.ndarray, y: np.ndarray, atol: float=1e-8) -> bool:
    """allclose is numba version of numpy.allcose.
    Returns True if two arrays are element-wise equal within a tolerance.

    Parameters
    ----------
    x,y : np.ndarray(3,)
        input vectors to compare
    atol : float, optional
        tolerance, by default 1e-8

    Returns
    -------
    bool
        True if vector has integer coordinates, False otherwise
    """    
    return np.all(np.abs(y-x)<=(atol+1e-5*np.abs(x)))

@njit()
def is_integer(x: np.ndarray, atol: float=1e-8) -> bool:
    """is_integer check if vector has integer coordinates within tolerance atol

    Parameters
    ----------
    x : np.ndarray(3,)
        input vector
    atol : float, optional
        tolerance, by default 1e-8

    Returns
    -------
    bool
        True if vector has integer coordinates, False otherwise
    """
    int_x = np.empty_like(x)
    np.round(x, 0, int_x)
    return allclose(x, int_x)

@njit()
def issamek(k1: np.ndarray, k2: np.ndarray, atol: float=1e-8) -> bool:
    """issamek checks if two k-points are equal modulo a lattice vector.

    Parameters
    ----------
    k1 : np.ndarray(3,)
        first k-point
    k2 : np.ndarray(3,)
        second k-point
    atol : float, optional
        tolerance, by default 1e-8

    Returns
    -------
    bool
        True if two k-points are the same
    """
    k1 = np.asarray(k1)
    k2 = np.asarray(k2)
    return is_integer(k1 - k2, atol=atol)

@njit()
def get_kq2k(kpts: np.ndarray, kpts_sub: np.ndarray, qpt: np.ndarray, atol_kdiff=1e-8) -> dict:
    """get_kq2k Find idx{k+q} --> idx{k} for k-point subset

    Parameters
    ----------
    kpts : np.ndarray(nk,3)
        array of k-points in the whole BZ
    kpts_sub : np.ndarray(nk_sub,3)
        array of k-point subset
    qpt : np.ndarray(3,)
        q-point for which we map BZ
    atol : float, optional
        tolerance, by default 1e-8
        
    Returns
    -------
    {int: (int, np.ndarray(3,))}
        dictionary which keys are k-point ind, values[0] is k+q ind, values[1] is G vector
    """
    k2kqg = dict()
    if np.all(np.abs(qpt) <= 1e-6):
        g0 = np.zeros(3)
        for ik, _ in enumerate(kpts_sub):
            k2kqg[ik] = (ik, g0)
    else:
        for ik, kpoint in enumerate(kpts_sub):
            kqpt = kpoint + qpt
            for ikq, kpt in enumerate(kpts):
                if issamek(kqpt, kpt, atol=atol_kdiff):
                    g0 = np.rint(kqpt - kpt)
                    k2kqg[ik] = (ikq, g0)
                    break
    return k2kqg

def get_eigvals_bz(eigvals_ibz: np.ndarray, bz_ibz_mapping: np.ndarray) -> np.ndarray:
    """get_eigvals_bz restores eigenvalues in full Brillouin zone from irreducible ones

    Parameters
    ----------
    eigvals_ibz : np.ndarray
        eigenvalues in IBZ
    bz_ibz_mapping : np.ndarray
        mapping of k-points from BZ to IBZ

    Returns
    -------
    np.ndarray
        eigenvalues in full BZ
    """    
    eigvals_bz = np.array([eigvals_ibz[i-1] for i in bz_ibz_mapping[:,0]])
    return eigvals_bz